/*******************************************************************************
*   ncap.c - Nagios Collector and Proxy 
*   Copyright (c) 2004 Giray Devlet <giray@osc.nl> 
*   License: GPL * * Last Modified: 2004-04-16
*
*   Command line: 
*          ncap -c <config_file> [-D]
*   
*   Description: 
*          This program is designed to run as a  background process and 
*          handles incoming requests (from the host running Nagios) to 
*          proxy plugin execution requests to ncap daemons.  
*
*   Initial code borrowed from  Ethan Galstad (nagios@nagios.org)'s nrpe 
******************************************************************************/


#define	_GNU_SOURCE 1         /*FIXME - probably breaks compatability */
#include "common.h"
#include "config.h"
#include "ncap.h"
#include "utils.h"
#include <time.h>
#include <stdio.h>
#include <pthread.h>

#ifdef HAVE_SSL
#include "ssl_thread_safe.h"
#include "dh.h"
#include <openssl/err.h>
#include <openssl/ssl.h>

#ifndef WIN32
#include <pthread.h>
#define THREAD_CC
#define THREAD_TYPE                    pthread_t
#define THREAD_CREATE(tid, entry, arg) pthread_create(&(tid), NULL, \
                                                          (entry), (arg))
#endif /* ifndef WIN32 */

#ifdef WIN32
#include <windows.h>
#define THREAD_CC                      __cdecl
#define THREAD_TYPE                    DWORD
#define THREAD_CREATE(tid, entry, arg) do { _beginthread((entry), 0, (arg));\
                                              (tid) = GetCurrentThreadId();   \
                                          } while (0)
#endif /* ifdef WIN32 */
#endif /* HAVE_SSL */

#define NASTY_METACHARS         "|`&><'\"\\[]{}"
#define DEFAULT_NRPE_COMMAND	"_NRPE_CHECK"   /* check version of NRPE daemon */

/* Log priority definitons */
#define NCAP_LOG_ERR      0     /* Errors */
#define NCAP_LOG_INFO     1     /* Informative messages */
#define NCAP_LOG_VERBOSE  2     /* Some verbosity, useless information */
#define NCAP_LOG_DEBUG    4     /* Real debug stuff */
#define NCAP_LOG_DEVEL    8     /* Developer stuff, very verbose */
#define NCAP_LOG_JUNK     16    /* Logs that even the developer hardly wants */

/* ----------------------------------------------- */
/*             Function Declarations               */ 
/* ----------------------------------------------- */

int process_arguments (int, char **);     /* Process command line arguments */
void display_version(void);               /* Show version information */
void display_usage(void);                 /* Show how to run programm */
void wait_for_connections (void);         /* NON FORKing ... Loop and wait for new connections */
void normalize_config_file_name(char *);  /* make sure the config file uses an absolute path */
int read_config_file (char *);            /* Process configuration file (ncap.cfg) */
int read_config_dir (char *);             /* Process configuration directories? */
int add_command (char *, char *, char *); /* Add a new command to linked-list */
command *find_command (char *, char *);   /* Search for a command in linked-list */
void sighandler (int);                    /* Signal Handler */
int drop_privileges (char *, char *);     /* Set new non-root UID and GID */
int check_privileges (void);              /* Check if we are running as root */
void free_memory (void);                  /* Clean up used memory */
int is_an_allowed_host (char *);          /* Check if incomming host is authorized to access us */
char *validate_request (ncap_packet *);   /* Sanitiy checking for incomming packet */
char *get_command_name (char *);          /* returns the command name out of the buffer*/
char *get_arguments (char *);             /* returns command arguments from the buffer*/
int contains_nasty_metachars (char *);    /* Check for dangerous META-Chars */
void alarm_handler (int sig);             /* used for sending nrpe information */
int check_nrpe (char *, int, char *, int, char *);      /* used for sending nrpe information */
int update_list (void);                   /* Process list and update information */
void *handle_connection (void *);         /* Handle incomming connection from check_ncap */
void *ncap_scheduler ();                  /* Scheduler thread, run through command list */
int ncap_log (int, ...);                  /* Generalised log function */


/* ----------------------------------------------- */
/*              Global Variables                   */
/* ----------------------------------------------- */

char config_file[MAX_INPUT_BUFFER] = "ncap.cfg";  /* Name of configuration file */
char allowed_hosts[MAX_INPUT_BUFFER] = " ";   /* List of hosts that are allowed to contact us */
int server_port = DEFAULT_SERVER_PORT;  /* The port we are listening at */
int client_port = DEFAULT_CLIENT_PORT;  /* The port nrpe is goint to listen at the other side */
char server_address[16] = "0.0.0.0";    /* Our IP address */
char client_address[16] = "0.0.0.0";    /* IP address of NRPE clients */
int socket_timeout = DEFAULT_SOCKET_TIMEOUT;     /* Time to wait for socket connection */
int command_counter = 0;                /* Counter keeping track of how many commands are there - debug */

command *command_list = NULL;           /* Linked list of commands */
pthread_mutex_t cmdlist_mutex = PTHREAD_MUTEX_INITIALIZER;

char *ncap_user = NULL;                 /* Which user are we running as */
char *ncap_group = NULL;                /* Which group are we running as */

int show_help = FALSE;                  /* Command Line - Show Help */
int show_license = FALSE;               /* Command Line - Show License */
int show_version = FALSE;               /* Command Line - Show Version */
int debug = TRUE;                       /* Debugging messages on? */
unsigned log_level = 0;                 /* Level of verbosity, aimed to replace debug */
int devel_debug = FALSE;                /* Developer Debugging messages on? */
int transparent_proxy = TRUE;           /* are we running in transparent proxy mode? */
int daemonize = TRUE;                   /* fork() and become a daemon process */
int old_age = OLD_AGE;                  /* Seconds after which a command will be dropped */
unsigned scheduler_wait = SCHEDULER_WAIT;    /* Amount of time to wait between relays of commands */     

#ifdef HAVE_SSL
SSL_METHOD *meth = NULL;    /* for listener thread */
SSL_CTX *ctx = NULL;        /* for listener thread */
SSL_METHOD *meth2 = NULL;   /* for check_ncap */
SSL_CTX *ctx2 = NULL;       /* for check_ncap */
int use_ssl = TRUE;
#else
int use_ssl = FALSE;
#endif



/*  ---------------------------------------------------------------------
 *
 * main()
 *
 --------------------------------------------------------------------- */
int
main (int argc, char **argv)
{
  int result = OK;
  int pid = 0;
#ifdef HAVE_SSL
  DH *dh = NULL;
#endif

  result = process_arguments (argc, argv);
  debug = 1; /* FIXME */

  if (result != OK || show_help == TRUE || show_license == TRUE
      || show_version == TRUE) {
      display_version();
  }

  if (show_license == TRUE) {
    display_license();
  } else if (result != OK || show_help == TRUE) {
    display_usage();
  }

  if (result != OK || show_help == TRUE || show_license == TRUE
      || show_version == TRUE) {
    exit (STATE_UNKNOWN);
  }

  if (debug) fprintf(stderr, "Starting up...\n"); /* FIXME */

  openlog ("ncap", LOG_PID, LOG_LOCAL2);

  /* make sure the config file uses an absolute path */
  normalize_config_file_name(config_file);

  result = read_config_file (config_file);

  /* exit if there are errors... */
  if (result == ERROR) {
    ncap_log (NCAP_LOG_ERR, 
              "Config file '%s' contained errors, bailing out...\n",
              config_file);
    return STATE_CRITICAL;
  }

  fprintf(stderr, "Config read, debug is %d...\n", debug); /* FIXME */

  generate_crc32_table();

#ifdef HAVE_SSL
  /* initialize SSL in main() */
  if (use_ssl == TRUE) {
    //SSL_library_init ();
    if (!THREAD_setup() || !SSL_library_init()){
      ncap_log (NCAP_LOG_ERR, "Fatal: Could not initialize OpenSSL.\n");
      exit(STATE_CRITICAL);
    }
    SSL_load_error_strings ();
    SSLeay_add_ssl_algorithms ();
    meth = SSLv23_server_method ();
    if ((ctx = SSL_CTX_new (meth)) == NULL) {
      ncap_log (NCAP_LOG_ERR, "Error: could not create SSL context.\n");
      exit (STATE_CRITICAL);
    }

    /* initialize SSL stuff for check_ncap */
    meth2 = SSLv23_client_method ();
    if ((ctx2 = SSL_CTX_new (meth2)) == NULL) {
      ncap_log(NCAP_LOG_ERR,
              "NCAP:CHECK_NRPE: Error - could not create SSL context.\n");
      exit (STATE_CRITICAL);
    }
    /*SSL_CTX_set_cipher_list(ctx,"ALL"); */
    SSL_CTX_set_cipher_list (ctx, "ADH");
    dh = get_dh512 ();
    SSL_CTX_set_tmp_dh (ctx, dh);
    DH_free (dh);


    ncap_log (NCAP_LOG_INFO,
              "INFO: SSL/TLS initialized. All network traffic will be encrypted.\n");
  }
  else {
    ncap_log(NCAP_LOG_INFO,
              "INFO: SSL/TLS NOT initialized. Network encryption DISABLED.\n");
  }
#endif

  ncap_log (NCAP_LOG_DEBUG, "Starting up the server process...\n");

  if (daemonize == TRUE) {
    pid = fork();
  } else {
    pid = 0;  /* force it to start the real work */
  }

  if (pid == 0) {
    THREAD_TYPE   scheduler_thread;
    ncap_log (NCAP_LOG_DEBUG, "Setting up daemon environment...\n");

    if (daemonize == TRUE) {
      /* we're a daemon - set up a new process group */
      setsid ();
      /* ignore SIGHUP */
      signal (SIGHUP, SIG_IGN);

      chdir ("/");
      umask (0);

      /* close standard file descriptors */
      close (0); close (1); close (2); 

      /* redirect standard descriptors to /dev/null */
      open ("/dev/null", O_RDONLY);
      open ("/dev/null", O_WRONLY);
      open ("/dev/null", O_WRONLY);
    }

    /* drop privileges */
    drop_privileges (ncap_user, ncap_group);

    /* make sure we're not root */
    check_privileges ();

    /* start the scheduler thread */
    if (THREAD_CREATE(scheduler_thread, (void *)ncap_scheduler, NULL) != 0){
      ncap_log(NCAP_LOG_ERR, "Error: could not start scheduler: %s\n", strerror(errno));
      exit (STATE_CRITICAL);
    }
    else {
      ncap_log(NCAP_LOG_DEBUG, "Scheduler started id = %lu.\n", 
                scheduler_thread);
    }

    
    ncap_log(NCAP_LOG_DEBUG, "Starting listener...\n");
    wait_for_connections ();
  }

#ifdef HAVE_SSL
  if (use_ssl == TRUE)
    SSL_CTX_free (ctx);
    SSL_CTX_free (ctx2);
#endif

  return STATE_OK;
} /* main() */

/* --------------------------------------------------------------------------- 
 * ncap_scheduler()
 *   This function implements the thread to walk through 
 *   the command list and maintain regular checks. 
 *   Reads/updates global `commad_list'.
 --------------------------------------------------------------------------- */
void *
ncap_scheduler()
{
time_t start_time = 0;  /* time stamp for scheduler start */
time_t stop_time  = 0;  /* time stape for scheduler stop  */
int    delta_time = 0;    /* time difference (stop-start)   */

  ncap_log(NCAP_LOG_VERBOSE, 
      "Scheduler started. Running through command_list every %d seconds.\n",scheduler_wait);
  while(1) {
    (void)time(&start_time);
    update_list();
    (void)time(&stop_time);
    delta_time=(int)stop_time-(int)start_time;
    ncap_log(NCAP_LOG_DEVEL, "command_list processing time: %d seconds.\n",delta_time);
    
    if(delta_time >= scheduler_wait){
      continue;
    } else {
      sleep ((scheduler_wait - delta_time));
    }
  }

} /* ncap_scheduler() */

/* ---------------------------------------------------------------------
 * display_version()
 *   show how the program is supposed to be run.
  --------------------------------------------------------------------- */

void
display_version(void)
{
    printf ("\n");
    printf ("NCAP - Nagios Collector and Proxy\n");
    printf ("Copyright (c) 2004 Giray Devlet <giray@osc.nl>\n");
    printf ("Version: %s\n", PROGRAM_VERSION);
    printf ("Last Modified: %s\n", MODIFICATION_DATE);
    printf ("License: GPL with exemptions (-l for more info)\n");
#ifdef HAVE_SSL
    printf
      ("SSL/TLS Available: Anonymous DH Mode, OpenSSL 0.9.6 or higher required\n");
#endif
    printf ("\n");
} /* display_version() */

/* ---------------------------------------------------------------------
 * display_usage()
 *   show how the program is supposed to be run.
  --------------------------------------------------------------------- */

void  
display_usage(void){
    printf ("Usage: ncap -c <config_file> \n\n");
    printf ("Options:\n");
    printf (" <config_file> = Name of config file to use\n\n");
    printf ("Notes:\n"
        "This program is designed to process requests from the check_ncap\n"
        "plugin on the host(s) running Nagios.\n"
        "Once a request is received from an authorized\n"
        "host, NCAP will proxy the command/plugin to a nrpe daemon.\n");
    return;
} /* display_usage() */

/* ---------------------------------------------------------------------
 * normalize_config_file_name()
 *   make sure the config file uses an absolute path 
  --------------------------------------------------------------------- */
void  
normalize_config_file_name(char * config_file)
{
  if (debug) fprintf(stderr, "Reading config ... %s\n",config_file); /* FIXME */
  char my_config_file[MAX_INPUT_BUFFER] = "";
  char buffer[MAX_INPUT_BUFFER] = "";

  strcpy(my_config_file, config_file);

  if (my_config_file[0] != '/') {

    /* save the name of the config file */
    strncpy (buffer, my_config_file, sizeof (buffer));
    buffer[sizeof (buffer) - 1] = '\x0';

    /* get absolute path of current working directory */
    strcpy (my_config_file, "");
    getcwd (my_config_file, sizeof (my_config_file));

    /* append a forward slash */
    strncat (my_config_file, "/", sizeof (my_config_file) - 2);
    my_config_file[sizeof (my_config_file) - 1] = '\x0';

    /* append the config file to the path */
    strncat (my_config_file, buffer,
             sizeof (my_config_file) - strlen (my_config_file) - 1);
    my_config_file[sizeof (my_config_file) - 1] = '\x0';
  }
  config_file=my_config_file;
  return;
} /* normalize_config_file_name() */

/* ---------------------------------------------------------------------
 * read_config_file()
 *   Reads in configuration file, and assigns
 *   configuration variables to program variables.
  --------------------------------------------------------------------- */

int
read_config_file (char *filename)
{
  FILE *fp = NULL;
  char config_file[MAX_FILENAME_LENGTH] = "";
  char input_buffer[MAX_INPUT_BUFFER] = "";
  char *temp_buffer = NULL;
  char *varname = NULL;
  char *varvalue = NULL;
  int line = 0;


  /* open the config file for reading */
  fp = fopen (filename, "r");

  /* exit if we couldn't open the config file */
  if (fp == NULL) {
    ncap_log(NCAP_LOG_ERR, "Unable to open config file '%s' for reading\n",
            filename);
    return ERROR;
  }

  line = 0;
  while (fgets (input_buffer, MAX_INPUT_BUFFER - 1, fp)) {

    line++;

    /* skip comments and blank lines */
    if (input_buffer[0] == '#')
      continue;
    if (input_buffer[0] == '\x0')
      continue;
    if (input_buffer[0] == '\n')
      continue;

    /* get the variable name */
    varname = strtok (input_buffer, "=");
    if (varname == NULL) {
      ncap_log(NCAP_LOG_ERR,
              "No variable name specified in config file '%s' - Line %d\n",
              filename, line);
      return ERROR;
    }

    /* get the variable value */
    varvalue = strtok (NULL, "\n");
    if (varvalue == NULL) {
      ncap_log(NCAP_LOG_ERR,
              "No variable value specified in config file '%s' - Line %d\n",
              filename, line);
      return ERROR;
    }

    /* allow users to specify directories to recurse into for config files */
    else if (!strcmp (varname, "include_dir")) {

      strncpy (config_file, varvalue, sizeof (config_file) - 1);
      config_file[sizeof (config_file) - 1] = '\x0';

      /* strip trailing / if necessary */
      if (config_file[strlen (config_file) - 1] == '/')
        config_file[strlen (config_file) - 1] = '\x0';

      /* process the config directory... */
      if (read_config_dir (config_file) == ERROR)
        break;
    }

    /* allow users to specify individual config files to include */
    else if (!strcmp (varname, "include")
             || !strcmp (varname, "include_file")) {
      if (read_config_file (varvalue) == ERROR) {    /* process the config file... */
        break;
      }
    } else if (!strcmp (varname, "server_port")) {
      server_port = atoi (varvalue);
      if (server_port < 1024) {
        ncap_log(NCAP_LOG_ERR,
                "Invalid port number specified (>= 1024) in config file '%s' - Line %d\n",
                filename, line);
        return ERROR;
      }
    } else if (!strcmp (varname, "client_port")) {
      client_port = atoi (varvalue);
      if (client_port < 1024) {
        ncap_log(NCAP_LOG_ERR,
                "Invalid port number specified (>= 1024) in config file '%s' - Line %d\n",
                filename, line);
        return ERROR;
      }
    } else if (!strcmp (varname, "server_address")) {
      strncpy (server_address, varvalue, sizeof (server_address) - 1);
      server_address[sizeof (server_address) - 1] = '\0';
    } else if (!strcmp (varname, "allowed_hosts")) {
      if (strlen (input_buffer) > sizeof (allowed_hosts)) {
        ncap_log(NCAP_LOG_ERR,
                "Allowed hosts list too long in config file '%s' - Line %d\n",
                filename, line);
        return ERROR;
      }
      strncpy (allowed_hosts, varvalue, sizeof (allowed_hosts));
      allowed_hosts[sizeof (allowed_hosts) - 1] = '\x0';
    } else if (strstr (input_buffer, "command[")) {
      temp_buffer = strtok (varname, "[");
      temp_buffer = strtok (NULL, "]");
      if (temp_buffer == NULL) {
        ncap_log(NCAP_LOG_ERR,
                "Invalid command specified in config file '%s' - Line %d\n",
                filename, line);
        return ERROR;
      }
      /* FIXME: Need to implement argument parsing from config file */
      add_command (temp_buffer, client_address, NULL);
    } else if (strstr (input_buffer, "log_level")) {
      log_level = (unsigned) strtol (varvalue, NULL, 0);
      if (log_level >= 4) {
        debug = TRUE;
        ncap_log(NCAP_LOG_DEBUG, "DEBUG set to TRUE\n");
      }
      else {
        debug = FALSE;
      }
       ncap_log(NCAP_LOG_INFO, "Debug Level set to %d",log_level); 
    } else if (strstr (input_buffer, "scheduler_wait")) {
      scheduler_wait = (unsigned) strtol (varvalue, NULL, 0);
    } else if (strstr (input_buffer, "transparent_proxy")) {
      transparent_proxy = atoi (varvalue);
      if (transparent_proxy > 0) {
        transparent_proxy = TRUE;
      }
      else {
        transparent_proxy = FALSE;
      }
    } else if (strstr (input_buffer, "old_age")) {
      old_age = atoi (varvalue);
      if (old_age < 10) {
        ncap_log(NCAP_LOG_ERR,
                "Less then 10 seconds?!?! This will never work - Line %d\n",
                filename, line);
      }
      else if (old_age < 300) {
        ncap_log(NCAP_LOG_ERR,
                "In a large environment you might run into problems with less then 5 minutes. - Line %d\n",
                filename, line);
      } /* if-then-else */

      ncap_log(NCAP_LOG_INFO,
              "Unused check_commands will be dropped after %d seconds.\n",
              old_age);
    } else if (!strcmp (varname, "ncap_user")) {
      ncap_user = strdup (varvalue);
    } else if (!strcmp (varname, "ncap_group")){
      ncap_group = strdup (varvalue);
    } else {
      ncap_log(NCAP_LOG_ERR,
              "Unknown option specified in config file '%s' - Line %d\n",
              filename, line);

      return ERROR;
    }
  } /* while */

  fclose (fp);  /* close the config file */

  return OK;
} /* read_config_file() */


/* ---------------------------------------------------------------------
 * read_config_dir()
 *    process all config files in a specific config 
 *    directory (with directory recursion) 
 --------------------------------------------------------------------- */

int
read_config_dir (char *dirname)
{
  char config_file[MAX_FILENAME_LENGTH];
  DIR *dirp = NULL;
  struct dirent *dirfile = {0};
  int result = OK;
  int x = 0;

  /* open the directory for reading */
  dirp = opendir (dirname);
  if (dirp == NULL) {
    ncap_log(NCAP_LOG_ERR, 
             "Could not open config directory '%s' for reading.\n", dirname);
    return ERROR;
  }

  /* process all files in the directory... */
  while ((dirfile = readdir (dirp)) != NULL) {

    /* process this if it's a config file... */
    x = strlen (dirfile->d_name);
    if (x > 4 && !strcmp (dirfile->d_name + (x - 4), ".cfg")) {

#ifdef _DIRENT_HAVE_D_TYPE
      /* only process normal files */
      if (dirfile->d_type != DT_REG)
        continue;
#endif

      /* create the full path to the config file */
      snprintf (config_file, sizeof (config_file) - 1, "%s/%s", dirname,
                dirfile->d_name);
      config_file[sizeof (config_file) - 1] = '\x0';

      /* process the config file */
      result = read_config_file (config_file);

      /* break out if we encountered an error */
      if (result == ERROR)
        break;
    }

#ifdef _DIRENT_HAVE_D_TYPE
    /* recurse into subdirectories... */
    if (dirfile->d_type == DT_DIR) {

      /* ignore current, parent and hidden directory entries */
      if (dirfile->d_name[0] == '.')
        continue;

      /* create the full path to the config directory */
      snprintf (config_file, sizeof (config_file) - 1, "%s/%s", dirname,
                dirfile->d_name);
      config_file[sizeof (config_file) - 1] = '\x0';

      /* process the config directory */
      result = read_config_dir (config_file);

      /* break out if we encountered an error */
      if (result == ERROR)
        break;
    }
#endif
  }

  closedir (dirp);

  return result;
} /* read_config_dir() */

/* ---------------------------------------------------------------------
 * add_command()
 *              adds a new command definition from the config file to 
 *              the list in memory.
 --------------------------------------------------------------------- */

int
add_command (char *command_name, char *client_address, char *arguments)
{
  command *new_command = NULL;
  time_t time_stamp = 0;

  (void) time (&time_stamp);
  /* allocate memory for the new command */
  new_command = (command *) malloc (sizeof (command));
  if (new_command == NULL) {
    ncap_log(NCAP_LOG_ERR, 
             "Failed adding command[%s] for client %s at %d: %s\n",
             command_name, client_address, time_stamp, strerror(errno));
    return ERROR;
  }

  strcpy (new_command->command_name, command_name);
  strcpy (new_command->client_address, client_address);
  strcpy (new_command->buffer, "NCAP has not processed this request yet.");
  new_command->result = STATE_UNKNOWN;
  new_command->time_stamp = (int) time_stamp;
  if (arguments != NULL) {
    strncpy (new_command->arguments, arguments, MAX_COMMANDNAME_LENGTH - 1);
  } else {
    strcpy (new_command->arguments,"");
  }
  command_counter++;

  /* add new command to head of list in memory */
  if(pthread_mutex_lock(&cmdlist_mutex) != 0) {
    ncap_log(NCAP_LOG_ERR, 
             "Locking cmdlist mutex failed in add_command(): %s\n",
              strerror(errno));
  }
  new_command->next = command_list;
  command_list = new_command;
  if(pthread_mutex_unlock(&cmdlist_mutex) != 0) {
    ncap_log(NCAP_LOG_ERR, 
             "Unlocking cmdlist mutex failed in add_command(): %s\n",
              strerror(errno));
  }

    ncap_log(NCAP_LOG_DEBUG,
            "Added command[%s][%s] for client %s with time %d command #%d\n",
            command_list->command_name, command_list->arguments, command_list->client_address,
            command_list->time_stamp, command_counter);

  return OK;
} /* add_command() */

/* ---------------------------------------------------------------------
 * find_command()
 *    given a command name, find the structure in memory
 * --------------------------------------------------------------------- */

command *
find_command (char *command_name, char *client_address)
{
  command *temp_command = NULL;
  time_t time_stamp = 0;

  (void) time (&time_stamp);
  if(pthread_mutex_lock(&cmdlist_mutex) != 0) {
    ncap_log(NCAP_LOG_ERR, 
             "Locking cmdlist mutex failed in find_command(): %s\n",
              strerror(errno));
  }
  for (temp_command = command_list; temp_command != NULL;
       temp_command = temp_command->next) {
    ncap_log(NCAP_LOG_JUNK, "#%d[%s][%s]@%d\n", command_counter,
            temp_command->command_name, temp_command->client_address,
            temp_command->time_stamp);
    if ((!strcmp (command_name, temp_command->command_name))
        && (!strcmp (client_address, temp_command->client_address))) {
      /* Update time stamp of accessed command */
      temp_command->time_stamp = (int) time_stamp;
      ncap_log(NCAP_LOG_JUNK, 
               "Found command[%s][%s] for client %s with time %d\n",
               temp_command->command_name, temp_command->arguments, temp_command->client_address,
               temp_command->time_stamp);
      if(pthread_mutex_unlock(&cmdlist_mutex) != 0) {
        ncap_log(NCAP_LOG_ERR, 
                "Unlocking cmdlist mutex failed in find_command(): %s\n",
                 strerror(errno));
      }
      return temp_command;
    }
  }                             /*for*/
  if(pthread_mutex_unlock(&cmdlist_mutex) != 0) {
    ncap_log(NCAP_LOG_ERR, 
             "Unlocking cmdlist mutex failed in find_command(): %s\n",
              strerror(errno));
  }

  return NULL;
} /* find_command() */

/* ---------------------------------------------------------------------
 * wait_for_connections() 
 *    wait for incoming connection requests.  This
 *    run via the main thread. If a valid connection 
 *    is established a new thread is is started which 
 *    will handle the connection.
 * --------------------------------------------------------------------- */

void
wait_for_connections (void)
{
  struct sockaddr_in myname = {0};
  struct sockaddr_in *nptr = NULL;
  struct sockaddr addr = {0};
  int rc =0 ;
  int sock =0;
  int new_sd = 0;
  socklen_t addrlen ={0};
  char connecting_host[16];
  int flag = 1;

  sock = socket (AF_INET, SOCK_STREAM, 0);
  if (sock < 0) { /* exit if we couldn't create the socket */
    ncap_log(NCAP_LOG_ERR, "Network server socket failure (%d: %s)\n", 
             errno, strerror (errno));
    exit (STATE_CRITICAL);
  }

  /* set the reuse address flag so we don't get errors when restarting */
  flag = 1;
  if (setsockopt
      (sock, SOL_SOCKET, SO_REUSEADDR, (char *) &flag, sizeof (flag)) < 0) {
    ncap_log(NCAP_LOG_ERR, "Could not set reuse address option on socket!\n");
    exit (STATE_UNKNOWN);
  }

  myname.sin_family = AF_INET;
  myname.sin_port = htons (server_port);
  bzero (&myname.sin_zero, 8);

  /* what address should we bind to? */
  if (!strlen (server_address)) {
    myname.sin_addr.s_addr = INADDR_ANY;
  }
  else if (!my_inet_aton (server_address, &myname.sin_addr)) {
    ncap_log(NCAP_LOG_ERR, "Server address is not a valid IP address\n");
    exit (STATE_CRITICAL);
  }

  /* bind the address to the Internet socket */
  if (bind (sock, (struct sockaddr *) &myname, sizeof (myname)) < 0) {
    ncap_log(NCAP_LOG_ERR, "Network server bind failure (%d: %s)\n", 
             errno, strerror (errno));
    exit (STATE_CRITICAL);
  }

  /* open the socket for listening */
  if (listen (sock, 5) < 0) {
    ncap_log(NCAP_LOG_ERR, "Network server listen failure (%d: %s)\n", 
             errno, strerror (errno));
    exit (STATE_CRITICAL);
  }

  ncap_log(NCAP_LOG_INFO, "Starting up daemon");

  /* Trap signals */
  signal (SIGQUIT, sighandler);
  signal (SIGTERM, sighandler);

  ncap_log(NCAP_LOG_DEBUG, "Listening for connections on port %d\n",
          htons (myname.sin_port));
  ncap_log(NCAP_LOG_DEBUG, "Allowing connections from: %s\n", allowed_hosts);

  ncap_log(NCAP_LOG_DEVEL, "Looping for accept()...\n");
  while (1) {
    ncap_log(NCAP_LOG_DEBUG, "Waiting for network connections...\n");
    new_sd = accept (sock, 0, 0);
    ncap_log(NCAP_LOG_DEVEL, "New connection, sock = %d\n", new_sd);

    if (new_sd < 0) { /* hey, there was an error... */
      switch (errno) {
        case EINTR:
        case EAGAIN:
             ncap_log(NCAP_LOG_ERR, "accept() failed (%d: %s).\n", 
                     errno, strerror (errno));
             continue;
             break;
        default:
             ncap_log(NCAP_LOG_ERR, "accept() failed (%d: %s). Terminating.\n", 
                     errno, strerror (errno));
             exit (STATE_CRITICAL); /* FIXME: Do we really want to terminate at this point?*/
             break;
      }

    }

    addrlen = sizeof (addr);
    rc = getpeername (new_sd, &addr, &addrlen);
    if (rc < 0) {
      ncap_log(NCAP_LOG_ERR,
              "Error: Network server getpeername() failure (%d: %s)\n",
              errno, strerror (errno));
      continue;
    }

    nptr = (struct sockaddr_in *) &addr;

    ncap_log(NCAP_LOG_DEBUG, "Connection from %s port %d\n",
            inet_ntoa (nptr->sin_addr), nptr->sin_port);

    /* is this is a blessed machine? */
    snprintf (connecting_host, sizeof (connecting_host), "%s",
              inet_ntoa (nptr->sin_addr));
    connecting_host[sizeof (connecting_host) - 1] = '\x0';

    if (!is_an_allowed_host (connecting_host)) {
      ncap_log(NCAP_LOG_ERR, "Host %s is not allowed to talk to us!\n",
              connecting_host);
      continue;
    }
    else {
      THREAD_TYPE  new_thread;
      int       rc;

      ncap_log(NCAP_LOG_DEBUG, "Host address checks out ok\n");

      /* Create connection handler (Worker) Thread */
      rc = THREAD_CREATE(new_thread, (void *)handle_connection,(void *) &new_sd ); 
      if (rc != 0) {
        ncap_log(NCAP_LOG_ERR, 
                "Error creating new connection handler thread: %s.\n",
                strerror(errno));
      } 
      
    }
  } /* while(1) */
} /*  wait_for_connections() */

/* ---------------------------------------------------------------------
 * handle_connection() 
 *   handles a client connection handed down by 
 *   wait_for_connections(). If this is a previously
 *   unknown command then it is added to command_list, 
 *   and a 'UNKNOWN' state is returned to the client.
 *   Otherwise, the last known information is returned.
 *
 *   This is a seperate thread.
 * --------------------------------------------------------------------- */

void *
handle_connection (void *arg)
{
  u_int32_t calculated_crc32 = 0;     /* SSL */
  command *temp_command = NULL;       /* Pointer to struct of relevant command */
  ncap_packet receive_packet = {0};   /* Packet received from check_ncap */
  ncap_packet send_packet = {0};      /* Packet sent to check_ncap */
  int bytes_to_send = 0;
  int bytes_to_recv = 0;
  char buffer[MAX_INPUT_BUFFER] = " ";
  int result = STATE_OK;              /* State returned to check_ncap */
  int rc = 0;
  int sock = 0;
  char *command_name = NULL;          /* Command name received from check_ncap */
  char *temp_command_name = NULL;     /* Struct pointer from received packet */
  char *arguments = NULL;             /* Arguments received from check_ncap */
#ifdef DEBUG
  FILE *errfp = NULL;
#endif
#ifdef HAVE_SSL
  SSL *ssl = NULL;
#endif

  /* Detach thread */
  if (pthread_detach(pthread_self()) != 0){
    ncap_log(NCAP_LOG_ERR, 
             "Error detaching thread: %s.\n", strerror(errno));
  } else {
  ncap_log(NCAP_LOG_DEVEL, 
        "Created new thread to handle the connection: id = %lu.\n",
         pthread_self());
  }


  sock = *((int*)arg);
  ncap_log(NCAP_LOG_DEVEL, "handle_connection[%lu] sock fd = %d...\n", 
           pthread_self(), sock);

#ifdef OLDSTUFF
  fcntl (sock, F_SETFL, O_NONBLOCK); /* socket should be non-blocking */
#endif

#ifdef HAVE_SSL
  /* do SSL handshake */
  if (result == STATE_OK && use_ssl == TRUE) {
    if ((ssl = SSL_new (ctx)) != NULL) {
      SSL_set_fd (ssl, sock);
      if ((rc = SSL_accept (ssl)) != 1) {
        char  error_buf[128];
        ncap_log(NCAP_LOG_ERR, "Error: Could not complete SSL handshake: %s\n",
                ERR_error_string (rc, error_buf));
#ifdef DEBUG
        errfp = fopen ("/tmp/err.log", "w");
        ERR_print_errors_fp (errfp);
        fclose (errfp);
#endif
        close(sock);
        ERR_remove_state(0);
        return NULL; 
      }
    }
    else {
      ncap_log(NCAP_LOG_ERR, 
               "Error: Could not create SSL connection structure.\n");
#ifdef DEBUG
      errfp = fopen ("/tmp/err.log", "w");
      ERR_print_errors_fp (errfp);
      fclose (errfp);
#endif
      close(sock);
      ERR_remove_state(0);
      return NULL;
    }
  }
#endif /* HAVE_SSL */

  bytes_to_recv = sizeof (receive_packet);
  if (use_ssl == FALSE) {
    rc = recvall (sock, (char *) &receive_packet, &bytes_to_recv,
                  socket_timeout);
  }
#ifdef HAVE_SSL
  else {
    rc = SSL_read (ssl, &receive_packet, bytes_to_recv);
  }
#endif

  /* recv() error or client disconnect */
  if (rc <= 0) {
#ifdef HAVE_SSL
    char error_buf[128];
#endif

    ncap_log(NCAP_LOG_ERR, 
             "Could not read request from client, bailing out...\n");

#ifdef HAVE_SSL
    ncap_log(NCAP_LOG_ERR, "SSL error: %s...\n", 
             ERR_error_string (rc, error_buf));
    if (SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN){
         SSL_shutdown (ssl);
    }
    SSL_free (ssl);
    ncap_log(NCAP_LOG_INFO, "INFO: SSL Socket Shutdown.\n");
#endif

    close(sock);
    return NULL;
  }
  /* we couldn't read the correct amount of data, so bail out */
  else if (bytes_to_recv != sizeof (receive_packet)) {
    ncap_log(NCAP_LOG_ERR, 
             "Data packet from client was too short, bailing out...\n");

#ifdef HAVE_SSL
    if (SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN){
         SSL_shutdown (ssl);
    }
    SSL_free (ssl);
#endif

    close(sock);
    ERR_remove_state(0);
    return NULL;
  }

#ifdef DEBUG
  fp = fopen ("/tmp/packet", "w");
  if (fp) {
    fwrite (&receive_packet, 1, sizeof (receive_packet), fp);
    fclose (fp);
  }
#endif

  /* make sure the request is valid */
  temp_command_name=validate_request (&receive_packet);
  command_name = get_command_name(temp_command_name);
  arguments    = get_arguments(temp_command_name);
  if ( command_name == NULL) {
    ncap_log(NCAP_LOG_ERR, "Client request was invalid, bailing out...\n");

    /* free memory */
    free (command_name);
    command_name = NULL;

#ifdef HAVE_SSL
    if (SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN){
         SSL_shutdown (ssl);
    }
    SSL_free (ssl);
#endif

    close(sock);
    ERR_remove_state(0);
    return NULL;
  }

  ncap_log(NCAP_LOG_DEBUG, 
           "Host is asking for command '%s' to be relayed ...\n",
           receive_packet.buffer);

  /* figure out what the nrpe client ip address is */
  strcpy (client_address, receive_packet.client_address);
  ncap_log(NCAP_LOG_DEBUG, "Target NRPE client is at %s.\n", client_address);

  /* if this is the version check command, just spew it out */
  if (!strcmp (command_name, NCAP_HELLO_COMMAND)) {

    snprintf (buffer, sizeof (buffer), "NCAP v%s", PROGRAM_VERSION);
    buffer[sizeof (buffer) - 1] = '\x0';

    ncap_log(NCAP_LOG_DEBUG, "Response: %s\n", buffer);

    result = STATE_OK;
  }
  else { /* find the command we're supposed to run */
    temp_command = find_command (command_name, client_address);

    if ((temp_command == NULL) && (!transparent_proxy)) {

      ncap_log(NCAP_LOG_DEBUG, "NCAP: Command '%s' for client '%s' not defined",
                command_name, client_address);
      snprintf (buffer, sizeof (buffer),
                "NCAP: Command '%s' for client '%s' not defined",
                command_name, client_address);
      buffer[sizeof (buffer) - 1] = '\x0';

      ncap_log(NCAP_LOG_DEBUG, "%s\n", buffer);
      result = STATE_UNKNOWN;
    } else {
      if (transparent_proxy == TRUE) { /* acting as transparent proxy */
        /* If we haven't encounter the command for this client before ... */
        if (temp_command == NULL) { 
            ncap_log(NCAP_LOG_DEBUG, "Previously unknown command for %s\n",
                    client_address);
          add_command (command_name, client_address, arguments);
          temp_command = find_command (command_name, client_address);
          strcpy (buffer, "NCAP: State UNKNOWN, scheduled for check");
          result = STATE_UNKNOWN;
        }
        else { /* we allready should have information about this host */
          /* Now we check if the arguments have changed */

          if (arguments == NULL) { 
             ncap_log(NCAP_LOG_ERR, "NCAP:************************************************");
             ncap_log(NCAP_LOG_ERR, "NCAP: ");
             ncap_log(NCAP_LOG_ERR, "NCAP: We ran out of memory and where about to crash");
             ncap_log(NCAP_LOG_ERR, "NCAP: So ... nothing more to do ... bye ");
             ncap_log(NCAP_LOG_ERR, "NCAP: ");
             ncap_log(NCAP_LOG_ERR, "NCAP:************************************************ ");
             exit(1);
          }

          if(strcmp(temp_command->arguments, arguments)){
             ncap_log(NCAP_LOG_DEBUG, "Argument list for command %s changed for host %s\n",
                        command_name, client_address);
 
             if(pthread_mutex_lock(&cmdlist_mutex) != 0) { /* lock command list */
               ncap_log(NCAP_LOG_ERR, 
                        "Locking cmdlist mutex failed in update_list(): %s\n",
                        strerror(errno));
             }
 
             /* change argument field in command list for this commnand */
             strncpy (temp_command->arguments, arguments, MAX_COMMANDNAME_LENGTH - 1);
             strcpy  (temp_command->buffer, "NCAP: State UNKNOWN, arguments changed. Scheduled for check");
 
             if(pthread_mutex_unlock(&cmdlist_mutex) != 0) { /* unlock command list */
               ncap_log(NCAP_LOG_ERR, 
                     "Unlocking cmdlist mutex failed in update_list(): %s\n",
                       strerror(errno));
             }
 
              strcpy (buffer, "NCAP: State UNKNOWN, arguments changed");
              result = STATE_UNKNOWN;
          }  else { /* Arguments have not changed */
               ncap_log(NCAP_LOG_DEVEL, "Arguments not changed\n");
               ncap_log(NCAP_LOG_DEBUG, 
                       "Found command in buffer %s\n", temp_command->buffer);
               strcpy (buffer, temp_command->buffer);
               result = temp_command->result;
          } 
          
        }
        ncap_log(NCAP_LOG_DEBUG, "Response: %s\n", buffer);

      }
      else { /* running in restrictive mode */
        ncap_log(NCAP_LOG_ERR,
            "NCAP does not support non-transparent proxy operations as yet\n");
        ncap_log(NCAP_LOG_ERR, "Check your configuration file ....\n");
        ncap_log(NCAP_LOG_ERR, "Good Bye ....\n");
        /* free all memory we allocated */
        free_memory ();
        /* terminate... */
        ERR_remove_state(0);
        exit (0);
        return NULL;
        /* FIXME: non-transparent (restrictive) proxy mode should be 
         * implemented for conrol freaks ...
         */
        }
      }
    }
  /* free memory */
  free (command_name);
  command_name = NULL;

  /* strip newline character from end of output buffer */
  if (buffer[strlen (buffer) - 1] == '\n')
    buffer[strlen (buffer) - 1] = '\x0';

  bzero (&send_packet, sizeof (send_packet));  /* clear the response packet buffer */

  /* fill the packet with semi-random data */
  //randomize_buffer ((char *) &send_packet, sizeof (send_packet));

  /* initialize response packet data */
  send_packet.packet_version = (int16_t) htons (NCAP_PACKET_VERSION_2);
  send_packet.packet_type = (int16_t) htons (RESPONSE_PACKET);
  send_packet.result_code = (int16_t) htons (result);
  strncpy (&send_packet.buffer[0], buffer, MAX_PACKETBUFFER_LENGTH);
  send_packet.buffer[MAX_PACKETBUFFER_LENGTH - 1] = '\x0';

  /* calculate the crc 32 value of the packet */
  send_packet.crc32_value = (u_int32_t) 0L;
  calculated_crc32 =
    calculate_crc32 ((char *) &send_packet, sizeof (send_packet));
  send_packet.crc32_value = (u_int32_t) htonl (calculated_crc32);

        /***** ENCRYPT RESPONSE *****/

  /* send the response back to the client */
  bytes_to_send = sizeof (send_packet);
  if (use_ssl == FALSE) {
    sendall (sock, (char *) &send_packet, &bytes_to_send);
  }
#ifdef HAVE_SSL
  else {
    SSL_write (ssl, &send_packet, bytes_to_send);
  }
#endif

#ifdef HAVE_SSL
  if (SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN){
       SSL_shutdown (ssl);
  }
  SSL_free (ssl);
#endif
  ncap_log(NCAP_LOG_DEBUG, "Return Code: %d, Output: %s\n", result, buffer);
  close(sock);
  ERR_remove_state(0);
  return NULL;
} /* handle_connection() */

/* --------------------------------------------------------------------- 
 * is_an_allowed_host()
 *   checks to see if a given host is 
 *   allowed to talk to us
 * --------------------------------------------------------------------- */

int
is_an_allowed_host (char *connecting_host)
{
  if (strstr(allowed_hosts, connecting_host) != NULL) {
    return 1;
  } else {
    return 0;
  }
} /* is_an_allowed_host() */

/* ---------------------------------------------------------------------
 * sighandler()
 *   handle signals 
 * --------------------------------------------------------------------- */

void
sighandler (int sig)
{
  /* free all memory we allocated */
  free_memory ();

  /* terminate... */
  exit (0);

  /* so the compiler doesn't complain.. */
  return;
} /* sighandler() */

/* ---------------------------------------------------------------------
 * free_memory()
 *   free all allocated memory 
 * --------------------------------------------------------------------- */

void
free_memory (void)
{
  command *this_command = {0};
  command *next_command = {0};

  /* free memory for the command list */
  this_command = command_list;
  while (this_command != NULL) {
    next_command = this_command->next;
    free (this_command);
    this_command = next_command;
  }

  return;
} /* free_memory() */

/* ---------------------------------------------------------------------
 * drop_privileges()
 *   drops privileges from our program. We should not 
 *   run as the user root.
 * --------------------------------------------------------------------- */

int
drop_privileges (char *user, char *group)
{
  uid_t uid = -1;
  gid_t gid = -1;
  struct group *grp = {0};
  struct passwd *pw = {0};
  
  if (group != NULL) { /* set effective group ID */
    /* see if this is a group name */
    if (strspn (group, "0123456789") < strlen (group)) {
      grp = (struct group *) getgrnam (group);
      if (grp != NULL) {
        gid = (gid_t) (grp->gr_gid);
      } else {
        ncap_log(NCAP_LOG_ERR, "Warning: Could not get group entry for '%s'\n",
                group);
      }
    } else { /* else we were passed the GID */
      gid = (gid_t) atoi (group);
    }

    /* set effective group ID if other than current EGID */
    if (gid != getegid ()) {
      if (setgid (gid) == -1){
        ncap_log(NCAP_LOG_ERR, "Warning: Could not set effective GID=%d\n",
                 (int) gid);
      }
    }
  } /* if(group != NULL) */

  
  if (user != NULL) { /* set effective user ID */
    /* see if this is a user name */
    if (strspn (user, "0123456789") < strlen (user)) {
      pw = (struct passwd *) getpwnam (user);
      if (pw != NULL) {
        uid = (uid_t) (pw->pw_uid);
      } else {
        ncap_log(NCAP_LOG_ERR, "Warning: Could not get passwd entry for '%s'\n",
                user);
      }
    }
  } else { /* else we were passed the UID */
      uid = (uid_t) atoi (user);
  } /* if (user != NULL) */

#ifdef HAVE_INITGROUPS
  if (uid != geteuid ()) {
    /* initialize supplementary groups */
    if (initgroups (user, gid) == -1) {
      if (errno == EPERM) {
        ncap_log(NCAP_LOG_ERR,
                "Warning: Unable to change supplementary groups using initgroups()\n");
      } else {
        ncap_log(NCAP_LOG_ERR,
                "Warning: Possibly root user failed dropping privileges with initgroups()\n");
        return ERROR;
      }
    }
  }
#endif

  if (setuid (uid) == -1){
    ncap_log(NCAP_LOG_ERR, 
             "Warning: Could not set effective UID=%d\n", (int) uid);
  }

  return OK;
} /* drop_privileges() */

/* ---------------------------------------------------------------------
 * check_privileges()
 *   bail if daemon is running as root
 * --------------------------------------------------------------------- */

int
check_privileges (void)
{
  uid_t uid = -1;
  gid_t gid = -1;

  uid = geteuid ();
  gid = getegid ();

  if (uid == 0 || gid == 0) {
    ncap_log(NCAP_LOG_ERR, 
             "Error: NCAP daemon cannot be run as user/group root!\n");
    exit (STATE_CRITICAL);
  }

  return OK;
} /* check_privileges() */

/* ---------------------------------------------------------------------
 * validate_request()
 *   tests whether or not a client request is valid 
 * --------------------------------------------------------------------- */

char *
validate_request (ncap_packet * pkt)
{
  u_int32_t packet_crc32 = 0;
  u_int32_t calculated_crc32 = 0;
  char *ptr = NULL;
  char *retval = NULL;

        /***** DECRYPT REQUEST ******/

  /* check the crc 32 value */
  packet_crc32 = ntohl (pkt->crc32_value);
  pkt->crc32_value = 0L;
  calculated_crc32 = calculate_crc32 ((char *) pkt, sizeof (ncap_packet));
  if (packet_crc32 != calculated_crc32) {
    ncap_log(NCAP_LOG_ERR, "Error: Request packet had invalid CRC32.\n");
    return NULL;
  }

  /* make sure this is the right type of packet */
  if (ntohs (pkt->packet_type) != QUERY_PACKET
      || ntohs (pkt->packet_version) != NCAP_PACKET_VERSION_2) {
    ncap_log(NCAP_LOG_ERR, "Error: Request packet type/version was invalid!\n");
    return NULL;
  }

  /* make sure buffer is terminated */
  pkt->buffer[MAX_PACKETBUFFER_LENGTH - 1] = '\x0';

  /* client must send some kind of request */
  if (!strcmp (pkt->buffer, "")) {
    ncap_log(NCAP_LOG_ERR, "Error: Request contained no query!\n");
    return NULL;
  }

  /* make sure request doesn't contain nasties */
  if (contains_nasty_metachars (pkt->buffer) == TRUE) {
    ncap_log(NCAP_LOG_ERR, "Error: Request contained illegal metachars!\n");
    return NULL;
  }

  retval = strdup(pkt->buffer);
  return retval;
} /* validate_request() */

/* ---------------------------------------------------------------------
 * get_command_name()
 *  returns the command name 
 * --------------------------------------------------------------------- */
char *
get_command_name(char * buffer){

  char *ptr = NULL;
  char *retval = NULL;
  char temp_buffer[MAX_INPUT_BUFFER] = "";

  /* get command name */
  if ((ptr = strchr(buffer, '!')) == NULL) {
    retval = strdup(buffer); /* there are no arguments to consider */
  } else {
    strcpy(temp_buffer,buffer);
    ptr=strtok(temp_buffer,"!");
    retval = strdup(ptr);
  }
  if (retval == NULL) {
    ncap_log(NCAP_LOG_ERR, "Error: Memory allocation failed in get_command_name()\n");
    return NULL;
  }
  return retval;

} /*get_command_name() */

/* ---------------------------------------------------------------------
 * get_arguments()
 *  returns the arguments 
 * --------------------------------------------------------------------- */
char *
get_arguments(char * buffer){

  char *ptr = NULL;
  char *retval = NULL;
  char empty_string[MAX_COMMANDNAME_LENGTH] = "";

  /* get command name */
  if ((ptr = strchr(buffer, '!')) == NULL) {
    retval = strdup(empty_string); /* there are no arguments */
    if (retval == NULL) {
      ncap_log(NCAP_LOG_ERR, "Error: (1) Memory allocation failed in get_arguments()\n");
      return NULL;
    }
  } else {
    retval = strdup(ptr);
    if (retval == NULL) {
      ncap_log(NCAP_LOG_ERR, "Error: (2) Memory allocation failed in get_arguments()\n");
      return NULL;
    }
  }
  return retval;

} /*get_arguments() */

/* ---------------------------------------------------------------------
 * contains_nasty_metachars()
 *   tests whether a buffer contains illegal metachars 
 * --------------------------------------------------------------------- */

int
contains_nasty_metachars (char *str)
{
  int result=0;

  if (str == NULL){
    return FALSE;
  }

  result = strcspn (str, NASTY_METACHARS);
  if (result != strlen (str)){
    return TRUE;
  }

  return FALSE;
} /* contains_nasty_metachars */

/* ---------------------------------------------------------------------
 * process_arguments()
 *   process command line arguments
 * --------------------------------------------------------------------- */

int
process_arguments (int argc, char **argv)
{
  char optchars[MAX_INPUT_BUFFER] = " ";
  int c = 1;

#ifdef HAVE_GETOPT_H
  int option_index = 0;
  static struct option long_options[] = {
    {"config", required_argument, 0, 'c'},
    {"no-ssl", no_argument, 0, 'n'},
    {"help", no_argument, 0, 'h'},
    {"license", no_argument, 0, 'l'},
    {"dont-fork", no_argument, 0, 'D'},
    {0, 0, 0, 0}
  };
#endif

  if (argc < 2){ /* no options were supplied */
    return ERROR;
  }

  snprintf (optchars, MAX_INPUT_BUFFER, "c:nhlD");

  while (1) {
#ifdef HAVE_GETOPT_H
    c = getopt_long (argc, argv, optchars, long_options, &option_index);
#else
    c = getopt (argc, argv, optchars);
#endif
    if (c == -1 || c == EOF) {
      break;
    }
    
    switch (c) { /* process all arguments */
      case '?':
      case 'h':
        show_help = TRUE;
        break;
      case 'V':
        show_version = TRUE;
        break;
      case 'l':
        show_license = TRUE;
        break;
      case 'c':
        strncpy (config_file, optarg, sizeof (config_file));
        config_file[sizeof (config_file) - 1] = '\x0';
        break;
      case 'n':
        use_ssl = FALSE;
        break;
      case 'D':
        daemonize = FALSE;
        break;
      default:
        return ERROR;
        break;
    } /* switch */
  }
  return STATE_OK;
} /* process_arguments() */

/* ---------------------------------------------------------------------
 * update_list()
 *   processes command_list by running trough every linked
 *   list node.  Each nodes command is sent to the nodes
 *   host.  The new information from the nrpe daemon is
 *   updated into the node.
 * --------------------------------------------------------------------- */

int
update_list ()
{

  command *temp_command = {0};        /* Pointer to current node in linked-list */
  command *prev_command = {0};        /* Pointer to previous node in linked-list */
  command *old_command = {0};         /* Pointer to now useless node in linked-list */
  int result = 0;
  time_t time_stamp = 0;
  char temp_buffer[MAX_INPUT_BUFFER] = "";


  (void) time (&time_stamp);    /* Get current time */

  if(command_list == NULL) {
     ncap_log(NCAP_LOG_DEBUG, "Command list empty.\n");
     return STATE_OK;
  }
  prev_command = NULL;
  if(pthread_mutex_lock(&cmdlist_mutex) != 0) {
    ncap_log(NCAP_LOG_ERR, 
             "Locking cmdlist mutex failed in update_list(): %s\n",
              strerror(errno));
  }
  temp_command = command_list;
  while(temp_command != NULL) {

    if ((time_stamp - temp_command->time_stamp) > old_age) {
      /* we are at the very beginning of the list */
      if (temp_command == command_list) {
        old_command = temp_command;
        ncap_log(NCAP_LOG_DEVEL, "temp: 0x%x, prev: 0x%x, list: 0x%x (1.0)\n",
                           temp_command, prev_command, command_list);
        if (devel_debug) {
          ncap_log(NCAP_LOG_DEBUG, "Command %s dropped out of old age for %s (1)\n",
                  old_command->command_name, old_command->client_address);
        }
        prev_command = NULL;
        temp_command = temp_command->next;
        command_list = temp_command;
        free (old_command);
        ncap_log(NCAP_LOG_DEVEL, "temp: 0x%x, prev: 0x%x, list: 0x%x (1.1)\n",
                           temp_command, prev_command, command_list);
      } else if (temp_command->next == NULL) { /* we are at the end of the list */
        old_command = temp_command;
        prev_command->next = NULL;
        if (devel_debug) {
          ncap_log(NCAP_LOG_DEBUG, "Command %s dropped out of old age for %s (2)\n",
                  old_command->command_name, old_command->client_address);
        }
        temp_command = NULL;
        free (old_command);
      } else { /* we're somewhere in the middle */
        old_command = temp_command;
        prev_command->next = temp_command->next;
        if (devel_debug) {
          ncap_log(NCAP_LOG_DEBUG, "Command %s dropped out of old age for %s (3)\n",
                  old_command->command_name, old_command->client_address);
        }
        temp_command = temp_command->next;
        free (old_command);
      }                         /* if-then-else */
      command_counter--;
    } else {
      if(pthread_mutex_unlock(&cmdlist_mutex) != 0) {
        ncap_log(NCAP_LOG_ERR, 
               "Unlocking cmdlist mutex failed in update_list(): %s\n",
                strerror(errno));
      }
      if (strlen(temp_command->arguments) > 1) { /* also pass arguments if they are there */
        strcpy(temp_buffer, temp_command->command_name);
        result = check_nrpe (temp_command->client_address,
                             client_port,
                             strcat(temp_buffer,temp_command->arguments),
                             TRUE, temp_command->buffer);
      } else { 
        result = check_nrpe (temp_command->client_address,
                             client_port,
                             temp_command->command_name,
                             TRUE, temp_command->buffer);
      }
      if(pthread_mutex_lock(&cmdlist_mutex) != 0) {
        ncap_log(NCAP_LOG_ERR, 
               "Locking cmdlist mutex failed in update_list(): %s\n",
                strerror(errno));
      }
      temp_command->result = result;
      ncap_log(NCAP_LOG_DEVEL, "Updated info for [%s][%s] with result=%d\n",
              temp_command->command_name, temp_command->arguments, temp_command->result);
      prev_command = temp_command;
      temp_command = temp_command->next;
      /* FIXME - what is temp_command->next if temp_command is NULL> */
    }                           /* if-then-else */
  } /* while */
  if(pthread_mutex_unlock(&cmdlist_mutex) != 0) {
    ncap_log(NCAP_LOG_ERR, 
           "Unlocking cmdlist mutex failed in update_list(): %s\n",
            strerror(errno));
  }
  return (STATE_OK);
} /* update_list() */

/* --------------------------------------------------------------------- 
 * ncap_log()
 *   log to syslog/terminal based on given log level 
 *   and debugging options.
 * --------------------------------------------------------------------- */

int 
ncap_log (int priority, ...) 
{
  va_list ap;
  char    log_buffer[1024] = "";
  char    *format = NULL;

  if (priority > log_level) return 0;

  va_start (ap, priority);
  format = va_arg(ap, char *);
  vsnprintf (log_buffer, 1023, format, ap);

  if(daemonize == TRUE) {
    switch (priority) {
      case NCAP_LOG_ERR: 
        syslog (LOG_ERR, log_buffer);
        break;
      case NCAP_LOG_INFO:
        syslog (LOG_INFO, log_buffer);
        break;
      default:
        syslog (LOG_DEBUG, log_buffer);
        break;
      }
  } else {
    fprintf(stderr, "%s", log_buffer);
  }
  va_end (ap);
  return 0;
} /* ncap_log() */

/* ---------------------------------------------------------------------
 * alarm_handler() 
 *   handles signals for network timeouts 
 * 
 * --------------------------------------------------------------------- */

void
alarm_handler (int sig)
{

  printf ("NCAP: CHECK_NRPE: Socket timeout after %d seconds.\n",
          socket_timeout);
  return;
} /* alarm_handler() */

/* ---------------------------------------------------------------------
 * check_nrpe() 
 *   Sliced and diced version of check_nrpe.c 
 *   (C)  Ethan Galstad (nagios@nagios.org)
 * --------------------------------------------------------------------- */

int
check_nrpe (char *nrpe_server_name, int nrpe_server_port, char *command_name,
            int use_ssl, char *result_buffer)
{

  int sd = 0;
  char query[MAX_INPUT_BUFFER] = "";
  strcpy (query, command_name);

#ifdef HAVE_SSL
  SSL *ssl = NULL;
#endif
  u_int32_t packet_crc32 = 0;
  u_int32_t calculated_crc32 = 0;
  int16_t result = 0;
  int rc = 0;
  nrpe_packet send_packet = {0};
  nrpe_packet receive_packet= {0};
  int bytes_to_send = 0;
  int bytes_to_recv = 0;

  /* generate the CRC 32 table */
  generate_crc32_table ();


  /* initialize alarm signal handling */
  signal (SIGALRM, alarm_handler);

  /* set socket timeout */
  alarm (socket_timeout);

  /* try to connect to the host at the given port number */
  result = my_tcp_connect (nrpe_server_name, nrpe_server_port, &sd);
  if(result != STATE_OK){
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Cannot connect to nrpe server: %s\n", strerror(errno));
      return STATE_CRITICAL;
  }

#ifdef HAVE_SSL
  /* do SSL handshake */
  if (result == STATE_OK && use_ssl == TRUE) {
    if ((ssl = SSL_new (ctx2)) != NULL) {
      SSL_CTX_set_cipher_list (ctx2, "ADH");
      SSL_set_fd (ssl, sd);
      if ((rc = SSL_connect (ssl)) != 1) {
        ncap_log(NCAP_LOG_ERR,
                "NCAP:CHECK_NRPE: Error - Could not complete SSL handshake.\n");
#ifdef DEBUG
        ncap_log(NCAP_LOG_ERR, "SSL_connect=%d\n", rc);
        /*
           rc=SSL_get_error(ssl,rc);
           syslog(LOG_ERR,"SSL_get_error=%d\n",rc);
           syslog(LOG_ERR,"ERR_get_error=%lu\n",ERR_get_error());
           syslog(LOG_ERR,"%s\n",ERR_error_string(rc,NULL));
         */
        ERR_print_errors_fp (stdout);
#endif
        result = STATE_CRITICAL;
      }
    }
    else {
      ncap_log(NCAP_LOG_ERR,
              "NCAP:CHECK_NRPE: Error - Could not create SSL connection structure.\n");
      result = STATE_CRITICAL;
    }

    /* bail if we had errors */
    if (result != STATE_OK) {
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Cannot connect to nrpe server.\n"); 
      return STATE_CRITICAL;
    }
  }
#endif

  /* we're connected and ready to go */
  if (result == STATE_OK) {

    /* clear the packet buffer */
    bzero (&send_packet, sizeof (send_packet));

    /* fill the packet with semi-random data */
    //randomize_buffer ((char *) &send_packet, sizeof (send_packet));

    /* initialize packet data */
    send_packet.packet_version = (int16_t) htons (NRPE_PACKET_VERSION_2);
    send_packet.packet_type = (int16_t) htons (QUERY_PACKET);
    strncpy (&send_packet.buffer[0], query, MAX_PACKETBUFFER_LENGTH);
    send_packet.buffer[MAX_PACKETBUFFER_LENGTH - 1] = '\x0';

    /* calculate the crc 32 value of the packet */
    send_packet.crc32_value = (u_int32_t) 0L;
    calculated_crc32 =
      calculate_crc32 ((char *) &send_packet, sizeof (send_packet));
    send_packet.crc32_value = (u_int32_t) htonl (calculated_crc32);


                /***** ENCRYPT REQUEST *****/


    /* send the packet */
    bytes_to_send = sizeof (send_packet);
    if (use_ssl == FALSE)
      rc = sendall (sd, (char *) &send_packet, &bytes_to_send);
#ifdef HAVE_SSL
    else {
      rc = SSL_write (ssl, &send_packet, bytes_to_send);
      if (rc < 0)
        rc = -1;
    }
#endif
    if (rc == -1) {
      ncap_log(NCAP_LOG_ERR, "NCAP:CHECK_NRPE: Error sending query to host.\n");
      close (sd);
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Cannot write to socket: %s.\n", strerror(errno)); 
      return STATE_UNKNOWN;
    }

    /* wait for the response packet */
    bytes_to_recv = sizeof (receive_packet);
    if (use_ssl == FALSE)
      rc =
        recvall (sd, (char *) &receive_packet, &bytes_to_recv,
                 socket_timeout);
#ifdef HAVE_SSL
    else
      rc = SSL_read (ssl, &receive_packet, bytes_to_recv);
#endif

    /* reset timeout */
    alarm (0);

    /* close the connection */
#ifdef HAVE_SSL
    if (use_ssl == TRUE) {
      if (SSL_get_shutdown(ssl) & SSL_RECEIVED_SHUTDOWN){
          SSL_shutdown (ssl);
      }
      SSL_free (ssl);
    }
#endif
    close (sd);

    /* recv() error */
    if (rc < 0) {
      ncap_log(NCAP_LOG_ERR,
              "NCAP:CHECK_NRPE: Error receiving data from daemon.\n");
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Error receiving data from daemon: %s.\n", strerror(errno)); 
      return STATE_CRITICAL;
    }
    else if (rc == 0) { /* nrpe_server disconnected */
      ncap_log(NCAP_LOG_ERR,
                "NCAP:CHECK_NRPE: Received 0 bytes from daemon."
                "Check the remote nrpe_server logs for error messages.\n");
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Error receiving data from daemon: %s.\n", strerror(errno)); 
      return STATE_CRITICAL;
    }
    else if (bytes_to_recv < sizeof (receive_packet)) { /* receive underflow */
      ncap_log(NCAP_LOG_ERR,
              "NCAP:CHECK_NRPE: Receive underflow - only %d bytes received"
              " (%d expected).\n", bytes_to_recv, sizeof (receive_packet));
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Error receiving data from daemon: %s.\n", strerror(errno)); 
      return STATE_CRITICAL;
    }


                /***** DECRYPT RESPONSE *****/


    /* check the crc 32 value */
    packet_crc32 = ntohl (receive_packet.crc32_value);
    receive_packet.crc32_value = 0L;
    calculated_crc32 =
      calculate_crc32 ((char *) &receive_packet, sizeof (receive_packet));
    if (packet_crc32 != calculated_crc32) {
      ncap_log(NCAP_LOG_ERR,
              "NCAP:CHECK_NRPE: Response packet had invalid CRC32.\n");
      close (sd);
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Response packet had invalid CRC32.\n"); 
      return STATE_UNKNOWN;
    }

    /* check packet version */
    if (ntohs (receive_packet.packet_version) != NRPE_PACKET_VERSION_2) {
      ncap_log(NCAP_LOG_ERR,
              "NCAP:CHECK_NRPE: Invalid packet version received from nrpe_server.\n");
      close (sd);
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Invalid packet version received from nrpe_server.\n"); 
      return STATE_UNKNOWN;
    }

    /* check packet type */
    if (ntohs (receive_packet.packet_type) != RESPONSE_PACKET) {
      ncap_log(NCAP_LOG_ERR,
              "NCAP:CHECK_NRPE: Invalid packet type received from nrpe_server.\n");
      close (sd);
      snprintf(result_buffer, MAX_PACKETBUFFER_LENGTH,
                 "Invalid packet type received from nrpe_server.\n"); 
      return STATE_UNKNOWN;
    }

    /* get the return code from the remote plugin */
    result = (int16_t) ntohs (receive_packet.result_code);

    /* print the output returned by the daemon */
    receive_packet.buffer[MAX_PACKETBUFFER_LENGTH - 1] = '\x0';
    if (!strcmp (receive_packet.buffer, "")) {
      ncap_log(NCAP_LOG_ERR,
              "NCAP:CHECK_NRPE: No output returned from daemon.\n");
      strcpy (result_buffer, receive_packet.buffer);
    }
    else {
      /* We got valid information */
      ncap_log(NCAP_LOG_DEBUG, "NCAP:CHECK_NRPE got: %s\n",
              receive_packet.buffer);
      /* return information to check_ncap / nagios */
      strcpy (result_buffer, receive_packet.buffer);
    }
  } else { /* reset the alarm */
    alarm (0);
  }
  return result;
} /* check_nrpe() */


/* eof */

/* vim: set ts=2 sw=2 et: */
